"""
TitanRunner.py
Launches Titan pipeline
Created on March 27 2014

@author: jrosner
"""

import argparse
import os
from ruffus import *
from helpers import runIDHelpers as rid
from helpers import parameters as params
from helpers import inputChecker as ic
from helpers import pipelineHelpers
from helpers import taskHelpers

__version__ = 'TitanRunner v0.1.0'

# main scope
if __name__ == '__main__':

    # command line arguments
    parser = argparse.ArgumentParser(description=__version__)
    parser.add_argument(
        '-i', '--infile', action='store', dest='infile', required=True,
        help='6 column (tab delimited) input file: tumourID   tumour_libID tumour_path'
        '   normalID    normal_libID normal_path')
    parser.add_argument(
        '--project-name', action='store', required=True, dest='project_name',
        help='name of the project')
    parser.add_argument(
        '--project-path', action='store', required=True, dest='project_path',
        help='specify results directory')
    parser.add_argument(
        '-c', '--configFile', action='store', required=True, dest='configfile',
        help='/path/to/config_file.cfg')
    parser.add_argument(
        '-r', '--run_id', action='store', type=int, dest='custom_run_id',
        help='manually enter run_id (used to re-launch a run)', default=False)
    parser.add_argument('--version', action='version', version=__version__)
    parser.add_argument(
        '--platform', default='illumina', choices=['illumina', 'solid'],
        dest='platform', help='set tech platform (default is illumina)')

    args = parser.parse_args()
    current_path = os.path.dirname(os.path.realpath(__file__))

    # write command line arguments to parameters module
    params.InitConfigReader(args.configfile)
    params.SetProjectName(args.project_name)
    params.SetProjectPath(args.project_path)
    params.SetReferenceVersion('hg19')
    params.SetReference('hg19')
    params.SetInfile(args.infile)
    params.SetPlatform(args.platform)
    params.SetProgramPath(current_path)

    # set the run id used for this run
    if args.custom_run_id == False:
        # increment and set runID
        rid.IncRunID(params.GetProjectName(), current_path)
    else:
        # use runid specified on the command line -r
        rid.SetRunID(args.custom_run_id, current_path)

        # check that the user specified run direcotry exists
        if not os.path.exists(taskHelpers.GetProjectNamePathRunID()):
            raise Exception('The project folder for this run does not exist')

    # ----------------------------------------------------------------------------#
    #  Launch Titan Pipeline
    # ----------------------------------------------------------------------------#
    from helpers import pipeline

    @follows(pipeline.CreateSegFiles, pipeline.PlotTitan)
    def LaunchPipeline():
        """initiates pipeline found in pipeline.py"""
        print 'TITAN COMPLETE'

    # log command line arguments
    pipelineHelpers.LogArguments(vars(args).items(), __version__)

    # log config file arguments
    pipelineHelpers.LogParameters('CLUSTER')
    pipelineHelpers.LogParameters('TITAN')

    # parse the input file
    infile_list = taskHelpers.ParseInfile(args.infile)
    line_count = len(infile_list)
    params.num_samples = len(infile_list)

    # log input bam file
    pipelineHelpers.LogInfile(infile_list)

    # verify inputs, print to screen and log
    (infile_pass, infile_msg) = ic.CheckInfile(args.infile)
    (config_pass, config_msg) = ic.CheckConfig(args.configfile)
    verify_msg = infile_msg + '\n' + config_msg
    pipelineHelpers.LogInputCheck(verify_msg)

    # terminate run if the inputChecker returns any failures
    if not (infile_pass and config_pass):
        raise Exception('Invalid input.')

    pipeline_msg = '\n---------------------------------\n' \
        'Running pipeline\n' \
        '---------------------------------'
    print pipeline_msg
    log = pipelineHelpers.GetLogFile('MAIN')
    pipelineHelpers.Logging('INFO', log, pipeline_msg)

    # LaunchPipeline()
    num_procs = line_count * 3
    pipeline_run([LaunchPipeline], multiprocess=num_procs, verbose=1)



