"""
pipeline.py
Contains the pipeline functions
Created on March 27 2014
@author: jrosner
"""

import os
import sys
from ruffus import *
from helpers import titanTasks, titanHelpers, pipelineHelpers
from helpers import runIDHelpers as rid
from helpers import parameters as params

current_path = params.GetProgramPath()
log = pipelineHelpers.GetLogFile('TITAN')


@files(titanTasks.ReadCountTaskList)
def ReadCounter(inputs, output_sentinel, outputs, sample_id, prev_sentinel):
    """Finds the counts for each bam file"""

    task_list = []
    log_msg = ' [ReadCounter] ' + '[' + sample_id + '] '

    pipelineHelpers.Logging('INFO', log, log_msg + 'Starting')
    if pipelineHelpers.CheckSentinel(prev_sentinel, log, log_msg):
        tissue_type = ['Tumour', 'Normal']
        read_counter = current_path + '/scripts/readCounter'
        python = sys.executable
        script_path = pipelineHelpers.GetScriptPath(
            sample_id, titanHelpers.name)
        titan_mem = titanHelpers.GetTitanMem('low')

        # Run read counter for each tumour normal pair
        for inp, outp, sample in zip(inputs[0], outputs[0], tissue_type):

            script = open(
                '{0}titan_{1}_readcount_{2}.sh'.format(script_path, sample_id,
                                                       sample), 'w')

            # Write GATK command to the script depending on the chromosome
            script.write('#!/bin/bash\n')
            script.write('{rc} -w 1000 -q 0 {bam} | '
                         '{py} {path}/scripts/filter_chromosomes.py > '
                         '{out}\n'.format(rc=read_counter, bam=inp,
                                          py=python, path=current_path, out=outp))
            script.close()

            process = pipelineHelpers.RunTask(
                os.path.abspath(script.name), 1, titan_mem,
                sample_id, titanHelpers.name)
            # Add task to tasklist
            task_list.append(process)

        # Checks which tasks are complete, and reruns tasks that have failed
        pipelineHelpers.CheckTaskStatus(
            task_list, output_sentinel, log, log_msg)
    pipelineHelpers.Logging('INFO', log, log_msg + 'Finished')


@follows(ReadCounter)
@files(titanTasks.CorrectReadsTaskList)
def CorrectReads(inputs, output_sentinel, outputs, sample_id, prev_sentinel):
    """applies correction to wig files"""

    log_msg = ' [CorrectReads] ' + '[' + sample_id + '] '

    if pipelineHelpers.CheckSentinel(prev_sentinel, log, log_msg):
        task_list = []
        pipelineHelpers.Logging('INFO', log, log_msg + 'Starting')

        # arguments
        titan_mem = titanHelpers.GetTitanMem('high')
        map_file, gc_file = titanHelpers.GetGCandMapFiles()

        # software and scripts
        r_exec = params.GetR()
        correct_reads_exec = current_path + '/scripts/correctReads.R'

        script_path = pipelineHelpers.GetScriptPath(
            sample_id, titanHelpers.name)
        full_id = sample_id + '_' + rid.GetRunID()
        #raise Exception(full_id)

        # Create script to run apolloh on the tumour heterozygous positions
        script = open(
            '{0}titan_{1}_titan_correctreads.sh'.format(script_path, sample_id), 'w')
        script.write('#!/bin/bash\n')
        script.write('{R} --no-save --args {samp} {wig1} {wig2} '
                     '{gcf} {mapf} {out} < {cr}'.format(R=r_exec, samp=full_id,
                                                        mapf=map_file, gcf=gc_file, wig1=inputs[
                                                        0][0], wig2=inputs[0][1],
                                                        out=outputs[0][0], cr=correct_reads_exec))
        script.close()

        # Adding script to cluster
        process = pipelineHelpers.RunTask(os.path.abspath(script.name), 1,
                                          titan_mem, sample_id, titanHelpers.name)
        task_list.append(process)

        # Checks which tasks are complete, and reruns tasks that have failed
        # creates a sentinel file after finishing regardless of success/failure
        pipelineHelpers.CheckTaskStatus(
            task_list, output_sentinel, log, log_msg)


@files(titanTasks.NormHetPosTaskList)
def NormalHetPositions(inputs, output_sentinel, output, sample_id, prev_sentinel):
    """Find all heterozygous positions from a normal sample"""

    log_msg = ' [NormalHetPositions] ' + '[' + sample_id + '] '
    pipelineHelpers.Logging('INFO', log, log_msg + 'Starting')

    if pipelineHelpers.CheckSentinel(prev_sentinel, log, log_msg):
        task_list = []

        # software
        samtools = params.GetSamtools()
        bcftools = params.GetBcftools()
        python = sys.executable

        # argumentss
        reference = params.GetReference()
        mapping_quality = titanHelpers.GetMappingQuality()
        base_quality = titanHelpers.GetBaseQuality()
        titan_mem = titanHelpers.GetTitanMem('med')

        # chromosome count
        chromosome_count = 1

        script_path = pipelineHelpers.GetScriptPath(
            sample_id, titanHelpers.name)

        # Adding a task to the cluster for each chromosome of the sample
        for out in output[0]:

            num = str(chromosome_count).zfill(2)
            script = open(
                '{0}titan_{1}_normalhetpsns_{2}.sh'.format(script_path,
                sample_id, num), 'w')

            # Write GATK command to the script depending on the chromosome
            script.write('#!/bin/bash\n')
            script.write('{st} view -b -F 0x0400 {inp} {ch} | {st} mpileup '
                         '-q {mq} -Q {bq} -uf {ref} - | {bcf} view -cgv - | {py} '
                         '{cp}/scripts/apollohFindHetPsns.py '
                         '{outf};'.format(st=samtools, inp=inputs[0][1],
                                          ch=pipelineHelpers.ConvertChromosome(
                                          chromosome_count),
                                          mq=mapping_quality, bq=base_quality, ref=reference,
                                          py=python, cp=current_path, outf=out, bcf=bcftools))

            script.close()

            # Adding script to the cluster
            process = pipelineHelpers.RunTask(
                os.path.abspath(script.name), 1, titan_mem,
                sample_id, titanHelpers.name)
            task_list.append(process)

            chromosome_count = chromosome_count + 1

        # Checks which tasks are complete, and reruns tasks that have failed,
        # creates a sentinel file after finishing regardless of success/failure
        pipelineHelpers.CheckTaskStatus(
            task_list, output_sentinel, log, log_msg)


@follows(NormalHetPositions)
@files(titanTasks.MergeNormChrTaskList)
def MergeNormalAllelicCounts(inputs, output_sentinel, output, sample_id, prev_sentinel):
    """Merge the normal allelic count files into one"""

    log_msg = ' [MergeNormalAllelicCounts] ' + '[' + sample_id + '] '
    pipelineHelpers.Logging('INFO', log, log_msg + 'Starting')

    if pipelineHelpers.CheckSentinel(prev_sentinel, log, log_msg):
        task_list = []

        script_path = pipelineHelpers.GetScriptPath(
            sample_id, titanHelpers.name)

        titan_mem = titanHelpers.GetTitanMem('med')

        script = open(
            '{0}titan_{1}_normmerge.sh'.format(script_path, sample_id), 'w')
        script.write('#!/bin/bash\n')

        # Merge each of the tumour chromosomes together
        script.write('cat ')
        for t_in in inputs[0]:
            script.write(t_in + ' ')
        script.write('> ' + output[0][0])
        script.close()

        process = pipelineHelpers.RunTask(
            os.path.abspath(script.name), 1, titan_mem,
            sample_id, titanHelpers.name)
        task_list.append(process)

        # Checks which tasks are complete, and reruns tasks that have failed,
        # creates a sentinel file after finishing regardless of success/failure
        pipelineHelpers.CheckTaskStatus(
            task_list, output_sentinel, log, log_msg)


@follows(NormalHetPositions)
@files(titanTasks.TumourAllelicCountTaskList)
def TumourAllelicCounts(inputs, output_sentinel, outputs, sample_id, prev_sentinel):
    """Find allelic counts for each position in the tumour sample  """

    log_msg = ' [TumourAllelicCounts] ' + '[' + sample_id + '] '

    if pipelineHelpers.CheckSentinel(prev_sentinel, log, log_msg):
        task_list = []
        pipelineHelpers.Logging('INFO', log, log_msg + 'Starting')

        chromosome_count = 1

        # parameters
        reference = params.GetReference()
        python = sys.executable
        mapping_quality = titanHelpers.GetMappingQuality()
        base_quality = titanHelpers.GetBaseQuality()
        script_path = pipelineHelpers.GetScriptPath(
            sample_id, titanHelpers.name)
        titan_mem = titanHelpers.GetTitanMem('med')

        # Assigning task for each chromosom of the sample
        for t_in, t_out in zip(inputs[1], outputs[0]):
            # Create the script file
            num = str(chromosome_count).zfill(2)
            script = open(
                '{0}titan_{1}_tumourcounts_{2}.sh'.format(script_path,
                                                          sample_id, num), 'w')

            # Find SNVs by writing samtools/SNVMuxsuite to script
            script.write('#!/bin/bash\n')
            script.write('{0} {1}/scripts/count.py {2} {3} {4} {5} {6} > '
                         '{7};'.format(python, current_path,
                                       t_in, inputs[0][0], reference,
                                       base_quality, mapping_quality, t_out))
            script.close()

            # Adding script to the cluster
            process = pipelineHelpers.RunTask(
                os.path.abspath(script.name), 1, titan_mem,
                sample_id, titanHelpers.name)
            task_list.append(process)

            chromosome_count = chromosome_count + 1

        # Checks which tasks are complete, and reruns tasks that have failed
        # creates a sentinel file after finishing regardless of success/failure
        pipelineHelpers.CheckTaskStatus(
            task_list, output_sentinel, log, log_msg)


@follows(TumourAllelicCounts)
@files(titanTasks.FilterTumourAllelicCountTaskList)
def FilterTumourAllelicCounts(inputs, output_sentinel, outputs, sample_id, prev_sentinel):
    """Filter counts based on whether they overlap a dbSNP position"""

    log_msg = ' [FilterTumourAllelicCounts] ' + '[' + sample_id + '] '

    if pipelineHelpers.CheckSentinel(prev_sentinel, log, log_msg):
        task_list = []
        pipelineHelpers.Logging('INFO', log, log_msg + 'Starting')
        chromosome_count = 1

        # get arguments
        python = sys.executable
        dbsnp_positions = titanHelpers.GetDbSnpRef()
        script_path = pipelineHelpers.GetScriptPath(
            sample_id, titanHelpers.name)
        titan_mem = titanHelpers.GetTitanMem('low')

        # Assigning task for each chromosome of the sample
        for t_in, t_out in zip(inputs[0], outputs[0]):

            num = str(chromosome_count).zfill(2)
            script = open(
                '{0}titan_{1}_filtercounts_{2}.sh'.format(script_path,
                                                          sample_id, num), 'w')

            script.write('#!/bin/bash\n')
            script.write('{py} {cwd}/scripts/filterCounts.py -t {tin} -r {ref} -c {chrom} '
                         '-o {tout};'.format(py=python, cwd=current_path,
                                             ref=dbsnp_positions, tin=t_in, tout=t_out,
                                             chrom=pipelineHelpers.ConvertChromosome(
                                                chromosome_count)))
            script.close()

            # Adding script to the cluster
            process = pipelineHelpers.RunTask(
                os.path.abspath(script.name), 1, titan_mem,
                sample_id, titanHelpers.name)
            task_list.append(process)

            chromosome_count += 1

        # Checks which tasks are complete, and reruns tasks that have failed
        # creates a sentinel file after finishing regardless of success/failure
        pipelineHelpers.CheckTaskStatus(
            task_list, output_sentinel, log, log_msg)


@follows(FilterTumourAllelicCounts)
@files(titanTasks.MergeTumChrTaskList)
def MergeTumourAllelicCounts(inputs, output_sentinel, output, sample_id, prev_sentinel):
    """Merge the het allelic count files into one"""

    log_msg = ' [MergeTumourAllelicCounts] ' + '[' + sample_id + '] '
    pipelineHelpers.Logging('INFO', log, log_msg + 'Starting')

    if pipelineHelpers.CheckSentinel(prev_sentinel, log, log_msg):
        task_list = []

        titan_mem = titanHelpers.GetTitanMem('med')

        # script path
        script_path = pipelineHelpers.GetScriptPath(
            sample_id, titanHelpers.name)

        script = open(
            '{0}titan_{1}_tummerge.sh'.format(script_path, sample_id), 'w')
        script.write('#!/bin/bash\n')

        # Merge each of the tumour chromosomes together
        script.write('cat ')
        for inp in inputs[0]:
            script.write(inp + ' ')
        script.write('> ' + output[0][0])
        script.close()

        process = pipelineHelpers.RunTask(
            os.path.abspath(script.name), 1, titan_mem,
            sample_id, titanHelpers.name)
        task_list.append(process)

        # Checks which tasks are complete, and reruns tasks that have failed
        # creates a sentinel file after finishing regardless of success/failure
        pipelineHelpers.CheckTaskStatus(
            task_list, output_sentinel, log, log_msg)


@follows(CorrectReads, MergeTumourAllelicCounts)
@files(titanTasks.RunTitanTaskList)
def RunTitan(inputs, output_sentinel, outputs, sample_id, prev_sentinel):
    """launches the titan algorithm in R"""

    log_msg = ' [RunTitan] ' + '[' + sample_id + '] '

    if pipelineHelpers.CheckSentinel(prev_sentinel, log, log_msg):
        task_list = []
        pipelineHelpers.Logging('INFO', log, log_msg + 'Starting')

        # parameters
        num_cores = titanHelpers.GetTitanCores()
        titan_ploidy = titanHelpers.GetTitanPloidy()
        titan_mem = titanHelpers.GetTitanMem('high')
        num_clusters = titanHelpers.GetNumClusters()
        map_file, gc_file = titanHelpers.GetGCandMapFiles()
        skew = titanHelpers.GetSkew()
        est_ploidy = titanHelpers.GetEstimatePloidy()
        norm_param_n0 = titanHelpers.GetNormParamNZero()
        est_method = titanHelpers.GetEstimateMethod()
        max_iters = titanHelpers.GetMaxIters()
        pseudo_counts = titanHelpers.GetPseudoCounts()
        txn_exp_len = titanHelpers.GetTxnExpLen()
        txn_z_strength = titanHelpers.GetTxnZStrength()
        alpha_high = titanHelpers.GetAlphaHigh()
        alpha_k = titanHelpers.GetAlphaK()
        max_cn = titanHelpers.GetMaxCN()
        map_file, gc_file = titanHelpers.GetGCandMapFiles()

        symmetric = 'TRUE'
        if params.GetPlatform() == 'solid':
            symmetric = 'FALSE'

        # software and scripts
        r_exec = params.GetR()
        titan_exec = current_path + '/scripts/titan.R'

        script_path = pipelineHelpers.GetScriptPath(
            sample_id, titanHelpers.name)
        full_id = sample_id + '_' + rid.GetRunID()

        for num_clust in range(int(num_clusters)):

            # Create script to run apolloh on the tumour heterozygous positions
            script = open(
                '{0}titan_{1}_titan_cluster0{2}.sh'.format(script_path, sample_id,
                                                           num_clust + 1), 'w')
            script.write('#!/bin/bash\n')
            script.write('{R} --no-save --args {samp} {tc_het} {cndata} {mapf} {n_clust} '
                         '{nc} {ploidy} {outf} {outp} {sk} {estP} {nzero} {estM} {mi} '
                         '{pscnts} {texp} {tzstr} {alph_k} {alph_h} {maxcn} {sym} < '
                         '{t_path}'.format(R=r_exec, samp=full_id, mapf=map_file, mi=max_iters,
                                           tc_het=inputs[0][0], sk=skew,
                                           outf=outputs[2 * num_clust][0],
                                           outp=outputs[2 * num_clust + 1][0],
                                           t_path=titan_exec, nc=num_cores,
                                           ploidy=titan_ploidy, n_clust=(num_clust + 1),
                                           estP=est_ploidy, nzero=norm_param_n0,
                                           cndata=inputs[1][0], estM=est_method,
                                           pscnts=pseudo_counts, texp=txn_exp_len,
                                           tzstr=txn_z_strength, alph_h=alpha_high,
                                           alph_k=alpha_k, maxcn=max_cn, sym=symmetric))
            script.close()
            # Adding script to cluster
            process = pipelineHelpers.RunTask(
                os.path.abspath(script.name), int(num_cores),
                titan_mem, sample_id, titanHelpers.name)
            task_list.append(process)

        # Checks which tasks are complete, and reruns tasks that have failed
        # creates a sentinel file after finishing regardless of success/failure
        pipelineHelpers.CheckTaskStatus(
            task_list, output_sentinel, log, log_msg)


@follows(RunTitan)
@files(titanTasks.PlotTitanTaskList)
def PlotTitan(inputs, output_sentinel, outputs, sample_id, prev_sentinel):
    """generates plots for all clusters"""

    log_msg = ' [PlotTitan] ' + '[' + sample_id + '] '

    if pipelineHelpers.CheckSentinel(prev_sentinel, log, log_msg):
        task_list = []
        pipelineHelpers.Logging('INFO', log, log_msg + 'Starting')

        num_clusters = titanHelpers.GetNumClusters()
        titan_mem = titanHelpers.GetTitanMem('high')
        # chroms = titanHelpers.GetChromosomes()

        # software and scripts
        r_exec = params.GetR()
        titan_plot_script = current_path + '/scripts/plot_titan.R'
        script_path = pipelineHelpers.GetScriptPath(
            sample_id, titanHelpers.name)

        for num_clust in range(int(num_clusters)):

            # Create script to run apolloh on the tumour heterozygous positions
            script = open(
                '{0}titan_{1}_plot_cluster_{2}.sh'.format(script_path,
                                                          sample_id, (num_clust + 1)), 'w')
            script.write('#!/bin/bash\n')
            script.write('{R} --no-save --args  {obj} {outd}'
                         ' {n_clust} {id} {runid} < {plot_path}'.format(R=r_exec, id=sample_id,
                                                                        runid=rid.GetRunID(),
                                                                        plot_path=titan_plot_script,
                                                                        obj=inputs[num_clust][0],
                                                                        outd=outputs[0][0],
                                                                        n_clust=(num_clust + 1)))
            script.close()
            # Adding script to cluster
            process = pipelineHelpers.RunTask(
                os.path.abspath(script.name), 1, titan_mem,
                sample_id, titanHelpers.name)
            task_list.append(process)

        # Checks which tasks are complete, and reruns tasks that have failed
        # creates a sentinel file after finishing regardless of success/failure
        pipelineHelpers.CheckTaskStatus(
            task_list, output_sentinel, log, log_msg)


@follows(RunTitan)
@files(titanTasks.CreateSegFilesTaskList)
def CreateSegFiles(inputs, output_sentinel, outputs, sample_id, prev_sentinel):
    """generates seg files for all clusters"""

    log_msg = ' [CreateSegFiles] ' + '[' + sample_id + '] '

    if pipelineHelpers.CheckSentinel(prev_sentinel, log, log_msg):
        task_list = []
        pipelineHelpers.Logging('INFO', log, log_msg + 'Starting')

        # parameters
        num_clusters = titanHelpers.GetNumClusters()
        titan_mem = titanHelpers.GetTitanMem('med')

        # software and scripts
        seg_script = current_path + '/scripts/createTITANsegmentfiles.pl'
        script_path = pipelineHelpers.GetScriptPath(
            sample_id, titanHelpers.name)

        for num_clust in range(int(num_clusters)):
            # Create script to run apolloh on the tumour heterozygous positions
            script = open('{0}titan_{1}_createSegs_cluster0{2}.sh'.format(
                script_path, sample_id, (num_clust + 1)), 'w')
            script.write('#!/bin/bash\n')
            script.write('{seg_scr} -id={samp} -infile={infile} -outfile={outf} '
                         '-outIGV={igv};'.format(samp=sample_id, infile=inputs[num_clust][0],
                                                 outf=outputs[2 * num_clust][0],
                                                 igv=outputs[2 * num_clust + 1][0],
                                                 seg_scr=seg_script))
            script.close()

            # Adding script to cluster
            process = pipelineHelpers.RunTask(
                os.path.abspath(script.name), 1, titan_mem, sample_id, titanHelpers.name)
            task_list.append(process)

        # Checks which tasks are complete, and reruns tasks that have failed
        # creates a sentinel file after finishing regardless of success/failure
        pipelineHelpers.CheckTaskStatus(
            task_list, output_sentinel, log, log_msg)
